/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package mindex.processors;

import messif.algorithms.AlgorithmMethodException;
import messif.objects.LocalAbstractObject;
import messif.objects.impl.ObjectFeature;
import messif.objects.impl.ObjectFeatureQuantized;
import messif.operations.AbstractOperation;
import messif.operations.Approximate;
import messif.operations.RankingSingleQueryOperation;
import messif.operations.query.QuantizeOperation;
import messif.operations.query.RangeQueryOperation;
import mindex.MIndexPPSimple;
import mindex.MetricIndex;
import mindex.navigation.MIndexLeafCell;

/**
 * Operation processor for QuantizeOperation with soft-assigment feature set on. It converts objects in operation 
 * into their quantized form (ObjectFeatureQuantized), where more keys can be assigned (soft-quantization).
 *
 * TODO: Base it on ApproxNavigationProcessor to identify "k" closest buckets?
 * 
 * @author Vlastislav Dohnal, Masaryk University, Brno, Czech Republic, dohnal@fi.muni.cz
 */
public class SoftQuantizeOperationNavigationProcessor extends ApproxNavigationProcessor {
    /** Type of quantization -- if values equals {@link #PARAM_TYPE_SOFT} this processor is used (soft-assignment), 
     * otherwise standard {@link QuantizeOperationNavigationProcessor} is used. */
    public static String PARAM_TYPE = "type";
    /** Value of {@link #PARAM_TYPE operation parameter TYPE} to use this soft-assignment */
    public static String PARAM_TYPE_SOFT = "soft";
    /** Limit of ratio on distances between pivots to stop selecting additional soft-assigned partitions */
    public static String PARAM_RATIO = "ratio";
    public static double PARAM_RATIO_DEFAULT = (double)0.9;
    /** Limit of distance of an object to its possible hyperplane between pivots to stop selecting additional soft-assigned partitions */
    public static String PARAM_DIST = "dist";
    /** Maximum number of additional soft-assigned partitions, i.e. additionial words count excluding the right/precise one. */
    public static String PARAM_MAX = "max";
    public static int PARAM_MAX_DEFAULT = 4;
    
    private final Type type;
    private final double limit;
    
    /** My own reference to the original operation since we replace it with RangeQueryOperation in the constructor */
    private final QuantizeOperation operation;

    /**
     * Creates new soft-quantization navigation processor for given operation and M-Index. The priority queue is initialized.
     * @param operation quantize operation
     * @param mIndex M-Index object with configuration and dynamic cell tree
     * @throws AlgorithmMethodException if the M-Index cannot determine the PP for query object
     */
    public SoftQuantizeOperationNavigationProcessor(QuantizeOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        super(new RangeQueryOperation(operation.getQueryObject(0), LocalAbstractObject.MAX_DISTANCE), new ApproximationSettings(operation), mIndex);
        this.operation = operation;
        if (operation.getParameter(PARAM_RATIO, Double.class) != null) {
            this.type = Type.RATIO;
            this.limit = operation.getParameter(PARAM_RATIO, Double.class, PARAM_RATIO_DEFAULT);
        } else {
            this.type = Type.DISTANCE;
            this.limit = operation.getParameter(PARAM_DIST, Double.class, 0d);
        }
    }
    
    /**
     * Process the approximate query on the data from given M-Index Voronoi leaf cell 
     *  (call standard {@link ApproxKNNQueryOperation#evaluate(messif.objects.util.AbstractObjectIterator) } on relevant data).
     * @param operation query operation to be processe on the leaf
     * @param leaf M-Index Voronoi leaf cell
     * @return number of objects processed (accessed) by the operation
     * @throws AlgorithmMethodException if the evaluate method fails
     */
    @Override
    protected int processOnLeaf(RankingSingleQueryOperation operation, MIndexLeafCell leaf, short[] leafPPP) throws AlgorithmMethodException {
        float[] dists = queryPPPDistCalculator.getQueryPivotDistances();
        short[] queryPPP = queryPPPDistCalculator.getQueryPPP();
        
        // Compare the ratios of pivot distance and check the threshold
        int len = Math.min(queryPPP.length, leafPPP.length);
        for (int i = 0; i < len; i++) {
            short pvtIn = queryPPP[i];
            short pvtOut = leafPPP[i];
            switch (type) {
                case RATIO:
                    double ratio = (pvtIn == pvtOut) ? 1.0 : (double)dists[pvtIn] / (double)dists[pvtOut];
                    if (ratio < limit)
                        return 0;
                    break;
                case DISTANCE:
                    if (Math.abs(dists[pvtOut] - dists[pvtIn]) > limit)
                        return 0;
                    break;
            }
        }
        // Add this PPP (cluster number) to the response.
        long[] quantizedKeys = new long[] { MIndexPPSimple.computeClusterNumber(leafPPP, mIndex.getNumberOfPivots(), 
                                                                                (short)leafPPP.length, mIndex.getMaxLevel()) };
        LocalAbstractObject quantObj;
        if (operation.getQueryObject() instanceof ObjectFeature) {
            ObjectFeature feat = (ObjectFeature)operation.getQueryObject();
            quantObj = new ObjectFeatureQuantized(feat.getX(), feat.getY(), feat.getOrientation(), feat.getScale(), quantizedKeys);
        } else {
            quantObj = new ObjectFeatureQuantized(0, 0, 0, 0, quantizedKeys);
        }
        quantObj.setObjectKey(operation.getQueryObject().getObjectKey());     // Set the original object key
        this.operation.addToAnswer(quantObj, this.operation.getAnswerCount(), null);    // Intead of distance use the ordinal number of the object
        return 1;
    }
 
    /**
     * Factory method for creating an instance of InsertOperationNavigationProcessor for given Bulk/InsertOperation
     * If allowed - initializing the objects (computing distances to pivots) is done in parallel. Acquires write lock on the leaf nodes.
     * @param operation - either instance of InsertOperation or BulkInsertOperation
     * @param mIndex
     * @return initialized instance of InsertOperationNavigationProcessor
     * @throws AlgorithmMethodException 
     */
    public static SoftQuantizeOperationNavigationProcessor getInstance(AbstractOperation operation, MetricIndex mIndex) throws AlgorithmMethodException {
        if (operation instanceof QuantizeOperation) {
            if (PARAM_TYPE_SOFT.equals(operation.getParameter(PARAM_TYPE, String.class))) {
                return new SoftQuantizeOperationNavigationProcessor((QuantizeOperation)operation, mIndex);
            }
            return null;
        } else {
            throw new UnsupportedOperationException("QuantizeOperationNavigationProcessor can be created only for QuantizeOperation");
        }
    }

    
    private static enum Type {
        RATIO, DISTANCE;
    }

    /** Class encapsulating the approximation settings */
    private static class ApproximationSettings implements Approximate {
        private int maxPartitionsAccessed;
        
        /**
         * Constructor with given search type and parameter.
         * @param localSearchType setter for {@link #localSearchType}
         * @param localSearchParam setter for {@link #localSearchParam} 
         */
        public ApproximationSettings(QuantizeOperation operation) {
            maxPartitionsAccessed = operation.getParameter(PARAM_MAX, Integer.class, PARAM_MAX_DEFAULT) + 1;
        }
        
        @Override
        public Approximate.LocalSearchType getLocalSearchType() {
            return Approximate.LocalSearchType.DATA_PARTITIONS;
        }

        @Override
        public int getLocalSearchParam() {
            return maxPartitionsAccessed;
        }

        @Override
        public float getRadiusGuaranteed() {
            return LocalAbstractObject.UNKNOWN_DISTANCE;
        }
      
        @Override
        public void setLocalSearchParam(int localSearchParam) {
        }

        @Override
        public void setLocalSearchType(Approximate.LocalSearchType localSearchType) {
        }

        @Override
        public void setRadiusGuaranteed(float radiusGuaranteed) {
        }

        @Override
        public String toString() {
            return "ApproximationSettings{" + "maxPartitionsAccessed=" + maxPartitionsAccessed + '}';
        }
    }
}
